%{
/*
 * tcsim.y - TC simulation language
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */


#include <stddef.h>
#include <stdlib.h>
#include <inttypes.h>

#include <netinet/in.h>


#define MAX_ARGC 4096
#define MAX_PACKET 70000


#include "tcsim.h"
#include "jiffies.h"
#include "host.h"
#include "var.h"
#include "command.h"


char *current_dev = NULL;

static int argc;
static char *argv[MAX_ARGC];
static unsigned char send_buf[MAX_PACKET];
static unsigned char *hex;
static struct host *current_host = NULL;
static char *echo_string = NULL;

static struct attributes attr;


static void add_echo_string(const char *next)
{
    int len;

    len = echo_string ? strlen(echo_string)+1 : 0;
    echo_string = realloc(echo_string,len+strlen(next)+1);
    if (!echo_string) {
	perror("realloc");
	exit(1);
    }
    if (len) echo_string[len-1] = ' ';
    strcpy(echo_string+len,next);
}

%}

%union {
    const char *str;
    uint32_t num;
    struct jiffval ujiff;
    struct command *cmd;
    struct net_device *dev;
};

%token		TOK_DEV TOK_TC TOK_TIME TOK_SEND TOK_POLL TOK_EVERY TOK_END
%token		TOK_HOST TOK_CONNECT TOK_ROUTE TOK_NETMASK TOK_DEFAULT
%token		TOK_INSMOD TOK_PRELOAD TOK_COMMAND
%token		TOK_NL TOK_TCC TOK_ECHO SHIFT_RIGHT SHIFT_LEFT
%token		TOK_NFMARK TOK_PRIORITY TOK_TC_INDEX
%token	<str>	TOK_WORD TOK_STRING ASSIGNMENT VARIABLE TOK_PRINTF_FORMAT
%token	<num>	TOK_NUM TOK_FORMAT TOK_DQUAD
%token	<ujiff>	TOK_UJIFFIES

%type	<num>	opt_expression opt_format opt_repetition
%type	<num>	expression inclusive_or_expression exclusive_or_expression
%type	<num>	and_expression shift_expression additive_expression
%type	<num>	multiplicative_expression unary_expression primary_expression
%type	<num>	opt_mask
%type	<ujiff>	delta_time
%type	<cmd>	command tc_command
%type	<dev>	opt_dev device
%type	<str>	printf_format

%%

all:
    assignments items
    ;

items:
    | items item
    ;

item:
    TOK_HOST '{'
	{
	    current_host = create_host();
	}
      host_items '}'
	{
	    current_host = NULL;
	}
    | connect assignments
    | device_def assignments
    | TOK_INSMOD TOK_WORD
	{
	    kernel_module($2);
	    free($2);
	}
    | TOK_PRELOAD TOK_WORD
	{
	    preload_tc_module($2);
	    free($2);
	}
    | TOK_TIME delta_time assignments
	{
	    if (terminating) yyerror("TIME after END");
	    if (advance_time($2) < 0) yyerror("can't turn back time");
	}
    | TOK_END assignments
	{
	    struct jiffval end = {
		.jiffies = ~0,
		.ujiffies = ~0,
	    };

	    terminating = 1;
	    (void) advance_time(end);
	}
    | command
	{
	    cmd_run($1);
	}
    | TOK_COMMAND ASSIGNMENT '=' command
	{
	    set_var_cmd($2,$4);
	    free($2);
	}
    ;

host_items:
    | host_item host_items
    ;

host_item:
    device_def
    | TOK_ROUTE route
    | connect
    | tc_command
	{
	    cmd_run($1);
	}
    ;

device_def:
    TOK_DEV TOK_WORD
	{
	    if (current_dev) free(current_dev);
	    current_dev = $2;
	}
      opt_expression opt_tcc
	{
	    create_net_device(current_host,$2,$4);
	}
    ;

route:
    TOK_DQUAD opt_mask device
	{
	    add_route(current_host,$1,$2,$3);
	}
    | TOK_DEFAULT device
	{
	    add_route(current_host,0,0,$2);
	}
    ;

opt_mask:
	{
	    $$ = 0xffffffff;
	}
    | TOK_NETMASK TOK_DQUAD
	{
	    $$ = $2;
	}
    ;

connect:
    TOK_CONNECT device device
	{
	    connect_dev($2,$3);
	}
    ;

opt_expression:
	{
	    $$ = -1;
	}
    | TOK_NUM
	{
	    $$ = $1;
	}
    ;

opt_tcc:
    | TOK_TCC
	/*
	 * TOK_TCC is for error checking only. Otherwise, one could have curly
	 * braces with tcc execution anywhere and the parser wouldn't notice
	 * until it's far too late.
	 */
    ;

argv:
	{
	    argv[argc] = NULL;
	}
    | argv TOK_WORD
	{
	    if (argc == MAX_ARGC) yyerror("too many arguments");
	    argv[argc++] = $2;
	}
    ;

command:
    tc_command
	{
	    $$ = $1;
	}
    | TOK_SEND opt_dev
	{
	   attr.nfmark = 0;
	   attr.priority = 0;
	   attr.tc_index = 0;
	}
      attributes assignments
	{
	    hex = send_buf;
	}
      hex_string
	{
	    $$ = cmd_send($2,send_buf,hex-send_buf,attr);
	}
    | TOK_POLL opt_dev assignments
	{
	    $$ = cmd_poll($2);
	}
    | TOK_EVERY delta_time command
	{
	    if (terminating) yyerror("EVERY doesn't work after END");
	    $$ = cmd_every($2,$3);
	}
    | TOK_ECHO echo_args assignments
	{
	    $$ = cmd_echo(echo_string);
	    echo_string = NULL;
	}
    | TOK_COMMAND VARIABLE assignments
	{
	    $$ = get_var_cmd($2);
	    free($2);
	}
    ;

attributes:
    | attributes attribute
    ;

attribute:
    TOK_NFMARK '=' expression
	{
	    attr.nfmark = $3;
	}
    | TOK_PRIORITY '=' expression
	{
	    attr.priority = $3;
	}
    | TOK_TC_INDEX '=' expression
	{
	    if ($3 > 0xffff)
		yyerrorf("tc_index %lu > 0xffff",(unsigned long) $3);
	    attr.tc_index = $3;
	}
    ;

echo_args:
    | echo_arg echo_args
    ;

echo_arg:
    TOK_STRING
	{
	    add_echo_string($1);
	    free($1);
	}
    | printf_format expression
	{
	    char tmp[30]; /* worst case: %u on 64 bit: 20 chars */

	    sprintf(tmp,$1 ? $1 : "%u",(unsigned int) $2);
	    if ($1) free($1);
	    add_echo_string(tmp);
	}
    ;

printf_format:
	{
	    $$ = NULL;
	}
    | TOK_PRINTF_FORMAT
	{
	    $$ = $1;
	}
    ;

tc_command:
    TOK_TC
	{
	    argv[0] = "tc";
	    argc = 1;
	}
      argv TOK_NL assignments
	{
	    int i;

	    $$ = cmd_tc(argc,argv);
	    for (i = 2; i < argc; i++) free(argv[i]);
	}
    ;

opt_dev:
	{
	    $$ = NULL;
	}
    | device
	{
	    $$ = $1;
	}
    ;

device:
    TOK_WORD
	{
	    $$ = lookup_net_device($1);
	    if (!$$) yyerrorf("unknown device \"%s\"",$1);
	    free($1);
	}
    ;

delta_time:
/*
 * Only allow typed time for now
 *
    TOK_NUM
	{
	    $$.jiffies = $1;
	    $$.ujiffies = 0;
	    if (!use_jiffies) $$ = stoj($$);
	}
    |
*/
    TOK_UJIFFIES
	{
	    $$ = $1;
	}
    ;

hex_string:
    | hex_value assignments hex_string
    ;

hex_value:
    opt_format expression opt_repetition
	{
	    int size,i;

	    size = ($1 < 0 ? -$1 : $1)/8;
	    if (!size) size = 1;
	    switch (size) {
		case 1:
		    if ($2 >= 0x100)
			yyerrorf("value 0x%lx is too big for a byte",
			  (unsigned long) $2);
		    break;
		case 2:
		    if ($2 >= 0x10000)
			yyerrorf("value 0x%lx is too big for a 16 bit word",
			  (unsigned long) $2);
		    break;
		default:
		    break;
	    }
	    for (i = 0; i < $3; i++) {
		if (hex+size > send_buf+MAX_PACKET) yyerror("packet too big");
		switch ($1) {
		    case 0:
		    case -8:
		    case 8:
			*hex++ = $2;
			break;
		    case -16:
			$2 = ntohs($2);
			/* fall through */
		    case 16:
			*hex++ = $2 >> 8;
			*hex++ = $2;
			break;
		    case -32:
			$2 = ntohl($2);
			/* fall through */
		    case 32:
			*hex++ = $2 >> 24;
			*hex++ = $2 >> 16;
			*hex++ = $2 >> 8;
			*hex++ = $2;
			break;
		}
	    }
	}
    ;

opt_format:
	{
	    $$ = 0;
	}
    | TOK_FORMAT
	{
	    $$ = $1;
	}
    ;

opt_repetition:
	{
	    $$ = 1;
	}
    | 'x' expression
	{
	    $$ = $2;
	}
    ;

assignments:
    | assignments ASSIGNMENT '=' expression
	{
	    set_var_unum($2,$4);
	    free($2);
	}
    ;

/*
 * Use this indirection so that we can add logical expressions in the future,
 * if we really want them ...
 */

expression:
    inclusive_or_expression
	{
	    $$ = $1;
	}
    ;

inclusive_or_expression:
    exclusive_or_expression
	{
	    $$ = $1;
	}
    | inclusive_or_expression '|' exclusive_or_expression
	{
	    $$ = $1 | $3;
	}
    ;

exclusive_or_expression:
    and_expression
	{
	    $$ = $1;
	}
    | exclusive_or_expression '^' and_expression
	{
	    $$ = $1 ^ $3;
	}
    ;

/*
 * Skip equality_expression and such, and go straight to shift_expression
 */

and_expression:
    shift_expression
	{
	    $$ = $1;
	}
    | and_expression '&' shift_expression
	{
	    $$ = $1 & $3;
	}
    ;

shift_expression:
    additive_expression
	{
	    $$ = $1;
	}
    | shift_expression SHIFT_RIGHT additive_expression
	{
	    $$ = $1 >> $3;
	}
    | shift_expression SHIFT_LEFT additive_expression
	{
	    $$ = $1 << $3;
	}
    ;

additive_expression:
    multiplicative_expression
	{
	    $$ = $1;
	}
    | additive_expression '+' multiplicative_expression
	{
	    $$ = $1+$3;
	}
    | additive_expression '-' multiplicative_expression
	{
	    $$ = $1-$3;
	}
    ;

multiplicative_expression:
    unary_expression
	{
	    $$ = $1;
	}
    | multiplicative_expression '*' unary_expression
	{
	    $$ = $1*$3;
	}
    | multiplicative_expression '/' unary_expression
	{
	    $$ = $1/$3;
	}
    | multiplicative_expression '%' unary_expression
	{
	    $$ = $1 % $3;
	}
   ;

unary_expression:
    primary_expression
	{
	    $$ = $1;
	}
    | '~' primary_expression
	{
	    $$ = ~$2;
	}
    ;

primary_expression:
    TOK_NUM
	{
	    $$ = $1;
	}
    | TOK_DQUAD
	{
	    $$ = $1;
	}
    | VARIABLE
	{
	    $$ = get_var_unum($1);
	}
    | '(' expression ')'
	{
	    $$ = $2;
	}
    ;
